# External Entities Development Notes

This file holds technical information on how the External Entities module is
architectured, how to us it programmatically and how to extend it through
plugins (ie. how to create plugins).


## Table of contents

- Glossary
- Architecture
- v2 to v3 changes
- v2 to v3 plugin upgrade
- Creating a storage client plugin
- Creating a field mapper plugin
- Creating a property mapper plugin
- Creating a data processor plugin
- Creating a data aggregator plugin
- Creating external entity types programmatically


## Glossary

Here is a set of "tentative" definitions for terms used in External Entities:
- Data aggregator: a plugin that handles common entity operations such has
  querying, loading, creating/saving, and deleting and manage a set of storage
  clients on the backend to perform those operations.
- Data processor: a plugin that gets a set of values as input and can apply
  altering operations on each value of that set to produce a new set of altered
  values that would fulfill a given use case. It may (or may not) be also able
  to revert the altering process to save values in their original format for
  instance.
- Entity: a Drupal entity (in a generic sense).
- External entity type: definition of an external entity with it fields, how
  they are mapped, its data sources (data aggregator, storage clients) and other
  settings such as if there are annotations and how caching is managed.
- External entity (instance): a Drupal entity which values are not stored into
  the Drupal database but rather in a external source (ie. "external" to Drupal
  but which could reside on the same server).
- Field: in a Drupal context, a Drupal field is an object that holds one or more
  property in a structured way to attach them to an entity. In a "source"
  context, it's just the key of an associative data array provided by a storage
  client.
- Field mapper: a plugin that provides a used interface to map a given field
  type to source data. The plugin usualy delegates the field property mapping to
  property mapper plugins but it may also completly handle the mapping by
  itself without property mappers.
- Original data: it is an associative array containing the original values
  provided by the "source" storage clients (after aggregation) and hold by the
  external entity.
- Property mapper: a plugin that can map source values to a given value (ie.
  a field property). A property mapper often uses a mapping expression provided
  by the user to extract the piece of data needed from the source values, but
  the "constant" property mapper is an example of exception to that rule.
  Property mappers often use a set of data processor to alter source values but
  it is possible, for a property mapper, to alter values by itself and not use
  data processors (ex.: xnttstrjp).
- Raw data: can be used to refer to original/source data as an associative
  array.
- Source: external data provider. Service or system that is queried and returns
  the values used to fill external entities.
- Storage client: a plugin that queries a source to get external data and that
  can also be used to send altered data to its source for saving. Common storage
  clients get their data from a single source which can be a REST service, a
  database, a file or a set of files.
- XNTT: it is a shortcut for "eXternal eNTiTy".


## Architecture

### Typical data stream
```
+----------------------+
| External storages    | 1  1 +----------------+ 1..* 1 +-----------------+
+======================+ <--> | Storage client | <----> | Data Aggregator |
| REST service,        |      +----------------+        +-----------------+
| files, databases,... |                                         ^ 1
+----------------------+                                         |
                                                                 v 1
                             +--------------+ *..0  1 +-----------------------+
                             | Field mapper | <-----> | ExternalEntityStorage |
                             +--------------+         +-----------------------+
                                     ^ 1                         ^ 1
                                     |                           |
                                     v 0..*                      v 1..*
                            +-----------------+         +-----------------+
                            | Property mapper |         | External entity |
                            +-----------------+         +-----------------+
                                     ^ 1
                                     |
                                     v 0..*
                             +----------------+
                             | Data processor |
                             +----------------+
```
@todo Describe the external storage to ExternalEntityStorage stream.
@todo Describe the ExternalEntityStorage to Data processor stream.
@todo Describe how external entities are queried, loaded, saved and deleted.


## v2 to v3 changes

v3 has been introduced to allow major External Entities upgrades (through
breaking changes) while keeping v2 available to let previous system continue
working without being required to upgrade. The v3 changes come from raised
issues, improvements and new feature requests, after several years of use of
External Entities in various contexts.

List of *BREAKING CHANGES* between External Entities v2 and v3:
@todo Update documentation. New features, behavior changes, syntax changes
(slash to dot for simple property mapper, pipe to equal sign for REST client
parameters).
!!! NOTE: THE FOLLOWING LIST NEEDS TO BE UPDATED !!!
- Global changes:
  - Type hinting has been added to v3 wherever possible.
  - ExternalEntityInterface and ExternalEntityTypeInterface have been moved to
    the Drupal\external_entities\Entity namespace to regroup them with their
    respective implementations.
  - Storage class name simplification: "ExternalEntityStorageClient*" become
    "StorageClient*".
  - Storage client plugin type name changed from "ExternalEntityStorageClient"
    to "StorageClient".
  - ExternalEntityGetMappableFieldsEvent event class does not have a field
    mapper anymore.
- Field Mapper
  - Removed unnecessary abstraction layers for field mappers:
    - "ExpressionFieldMapperInterface" has been merged to
      "FieldMapperInterface".
    - "ConfigurableExpressionFieldMapperBase" and "ExpressionFieldMapperBase"
      have beend merge to "FieldMapperBase".
    Those layers came from a discussion when introducing field mapper plugins:
    https://www.drupal.org/project/external_entities/issues/3079012#comment-13437754
    Field mapper plugins should now use "FieldMapperBase" class as base.
    The rational behind this change is that those layers were adding complexity
    to the code since some methods of "ExpressionFieldMapperInterface" were
    required at some places where "FieldMapperInterface" was used. It would be
    hard to choose where to use one interface or the other while only one is
    used by plugins and current external entities implementation forces the use
    of plugins for field mappers. If one day, someone would create external
    entities type from code and want to force the use of a non-plugin field
    mapper, either this person would have to turn it into a plugin (which would
    not be so complicated) or raise an issue with explanations on the use case.
  - "createRawDataFromEntityValues" has an additional parameter
    $original_values that should be the source array which will be overriden
    by the new values.
  - "getFieldPropertyMapping" has been removed and can be replaced by
    "getFieldMapping" as the possible return types would be the same now.
  - "getConstantMappingPrefix" has been removed as the mapping type (ie. for
    instance FieldMapperInterface::MAPPING_CONSTANT) is now provided with the
    mapping string when using "getFieldMapping".
  - "extractFieldValuesFromRawData" and "extractFieldPropertyValuesFromRawData"
    are now part of the interface and made public but are using the string field
    name as first argument instead of the field definition.
  - "getMappableFields" stores mappable fields in class member "mappableFields".
- REST client:
  - Method getListQueryParameters() does not have "$start" and "$length"
    parameters. It does not handle paging anymore as well: paging is managed by
    a modified getPagingQueryParameters() method.
  - Method getPagingQueryParameters() does not return an array of paging
    parameters but an array of paging parameters for multiple queries to map
    client paging to server paging (with different page sizes).
  - The REST client does not save data using "application/x-www-form-urlencoded"
    data by default anymore. It now uses the REST endpoint encoding type set by
    the "Endpoint data encoding format" setting.

List of v3 new features:
- Global:
  - New method to set and retrieve source raw data from an external entity:
    ´setSourceRawData(array)´ and ´getSourceRawData()´. When external entities
    are saved, source fields that were not mapped to Drupal fields are kept.
  - Plugins (Storage Clients and Field Mappers) can now rely on the method
    submitConfigurationForm() rather than validateConfigurationForm() to set
    their configuration changes while still use validateConfigurationForm() to
    validate/invalidate form submission.
  - Added support for debugging external entities (loading, mapping, saving).
  - Both field mapper and storage clients settings can have administrative notes
    for an easier maintenance (document specific setting reasons, other possible
    options not used, etc.).
  - External Entity Types can be created programatically or by config, with or
    without editing restrictions and are highly customizable (list builder,
    forms, etc.).
  - Plugins using PluginFormTrait now benefit from the new method
    ::buildAjaxParentSubForm() allowing to use Ajax call to update any parts of
    subforms.
- Storage clients:
  - New interfaces and base classes for categories of storage clients:
    - RestClientInterface and RestClient (previously just "Rest") to implement
      any custom REST storage client. Note: the "base" class RestClient can be
      used "as is" (ie. not an abstract class) and is therefore not suffixed
      "Base".
    - FileClientInterface and FileClientBase to implement any file storage
      client.
    - QueryLanguageClientInterface and QueryLanguageClientBase to implement any
      query language or database storage client.
- Aggregator plugin: a new aggregator plugin has been added to allow aggregation
  of multiple storage clients (similar to xnttmulti plugin used to do for v2).
- REST client:
  - New endpoint options to support specific URLs to fetch a single entity or to
    count entities.
  - Support for any response structure using JSONPath to locate data in result
    sets.
  - Improved pager settings to support server limits and paging starting from 1.
  - Better support for authentication. Bearer token built-in support. API key
    support through query string. Support for custom authentication header.
  - Added support for custom headers. Ex.: "Accept: application/vnd.api+json".
  - Additional parameters can now be sent either as a query string or as POST
    data using the endpoint data encoding format. It is possible to mix POST and
    GET parameters using a question mark to specify query parameters in the
    endpoint URL.
  - Added an entity identifier placeholder "{id}" replaced in single entity
    endpoint URL and parameters.
- Files storage client is part of external entities module.
  - The files storage plugin ID changed from 'xnttfiles' to just 'files'.


## v2 to v3 plugin upgrade

### Upgrading v2 storage clients
@todo Update documentation. Explain the Drupal-side and source-side differences.
!!! NOTE: THE FOLLOWING DOC NEEDS TO BE REVIEWED AND UPDATED !!!
- In the module's composer.json (if one) update requirements ("require") from
  `"drupal/external_entities": "^2.0@alpha"`
  (or similar) to
  `"drupal/external_entities": "^3.0@alpha"`
- In the storage class, update annotations from:
  ` * @ExternalEntityStorageClient(`
  to
  ` * @StorageClient(`
- Adjust ExternalEntityInterface and ExternalEntityTypeInterface namespaces.
- If the storage client implements the constructor (`__construct()`) and the
  `create()` methods, ResponseDecoderFactoryInterface has been removed and
  two other services are added: LoggerChannelFactoryInterface, and
  EntityFieldManagerInterface. Make the changes in both methods unless you use
  ResponseDecoderFactoryInterface. In that later case, you will need to do the
  changes in the __constructor() when calling parent class constructor and
  add LoggerChannelFactoryInterface, and EntityFieldManagerInterface to your
  __constructor() and create() methods. You can get a logger channel factory and
  an entity field manager from core services using:
  ```
  $container->get('logger.factory')
  $container->get('entity_field.manager')
  ```
  Don't forget to update code documentation and to add the import:
  ```
  use Drupal\Core\Logger\LoggerChannelFactoryInterface;
  use Drupal\Core\Entity\EntityFieldManagerInterface;
  ```
  For the explanations, ResponseDecoderFactoryInterface was only needed by REST
  clients and was removed from the StorageClientInterface. On the other side,
  EntityFieldManagerInterface was needed by the new features added to
  EntityFieldManagerInterface for query filtering. And the logger factory should
  be provided to all storage clients as they all would need to log things.
- To log messages, you can now use `$this->logger->notice(...)` or
  `$this->logger->warning(...)` or whatever log method you need.
- If the storage client extends the built-in REST client, you will have to take
  into account new configuration items added as well as form element changes.
  Also note the Rest storage client class has been renamed into "RestClient" for
  consistency (regarding other base clients).
- You should provide an implementation for `submitConfigurationForm()` and move
  some parts of `validateConfigurationForm()` into this new handler that should
  call `$this->setConfiguration($form_state->getValues());` in the end:
  ```
  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    ...
    $this->setConfiguration($form_state->getValues());
  }
  ```
- Rename query() and countQuery() into querySource() and countQuerySource()
  respectively. This should be sufficient unless you already tried to translate
  Drupal field names into mapped source field names in your code logic. In that
  case, you may need to do some more code adjustments.
- Implement transliterateDrupalFilters() method. To start, you can use the
  basic but unefficient implementation:
  ```
  public function transliterateDrupalFilters(
    array $parameters,
    array $context = []
  ) :array {
    // @todo: To really implement for better performances.
    return $this->transliterateDrupalFiltersAlter(
      ['source' => [], 'drupal' => $parameters],
      $parameters,
      $context
    );
  }
  ```
  When implementing, you may find helpfull to use
  $this->externalEntityType->getFieldMapper() and $this->entityFieldManager new
  class member as well as the getFieldDefinitions() and getFieldMappings() new
  methods. The storage client "knows" which operators are supported by its
  storage source (list of used operators: see StorageClientInterface::query()
  documentation and Drupal Query API documentation:
  https://www.drupal.org/docs/8/api/database-api/dynamic-queries/conditions and
  https://api.drupal.org/api/drupal/core%21lib%21Drupal%21Core%21Entity%21Query%21QueryInterface.php/function/QueryInterface%3A%3Acondition/10).
  Then, some operators may or may not be supported by some source fields. While
  the $parameters parameter containes filter on *Drupal* field names, it will be
  required to match those to source fields using the field mapper
  ($this->externalEntityType->getFieldMapper() and getFieldMappings() method).
  And to tell if a given source field can support a given operator, you may rely
  on the Drupal mapped field type given by the user to guess the source field
  type ($this->entityFieldManager and getFieldDefinitions() method).
- For derived REST clients:
  - You should get familiar with the new features and see if some may not change
    your approach.
  - Update buildConfigurationForm() to adjust elements that should be visible or
    hidden with pre-filled values. You may also adjust existing and new field
    properties.
  - If overriden, update getListQueryParameters() signature to:
    ```
    protected function getListQueryParameters(array $filters = []) {
    ```
    and replace `$parameters` by `$filters`.
    Also remove the use of getPagingQueryParameters() in that method as it is
    handled separatly now.
  - If overriden, update getPagingQueryParameters() logic according to the new
    behavior. Maybe this new implementation would not need to be overriden and
    you may just remove the override.
  - If overriden, update getSingleQueryParameters() signature to:
    ```
    protected function getSingleQueryParameters(int|string|null $id) :array {
    ```
  - If overriden, update getHttpHeaders() signature to:
    ```
    public function getHttpHeaders(
      string $request_type = 'list',
      string $endpoint = '',
      array $parameters = []
    ) :array {
    ```
- For derived File clients:
  - Now xnttfiles plugin is part of external entities. You need to remove the
    dependency in you .info.yml file:
    ```
    dependencies:
    - external_entities:external_entities
    - xnttfiles:xnttfiles
    ```
    becomes:
    ```
    dependencies:
    - external_entities:external_entities
    ```
    and your module composer.json needs to be updated as well to require only
    external_entities v3:
    ```
    "require": {
      "drupal/external_entities": "^2.0@beta",
      "drupal/xnttfiles": "^1.0@beta"
    }
    ```
    becomes:
    ```
    "require": {
      "drupal/external_entities": "^3.0@beta"
    }
    ```
  - Method ::getIdField() has been removed and can be replaced by base
    class method ::getSourceIdFieldName().
  - The internal raw field added to store file path has been renamed from
    '_xnttfiles_path' to '_xntt_path'.
  - The config structure changed a little (see
    FileClientBase::defaultConfiguration() and Files::defaultConfiguration()).
    You may need to implement a hook_update_N() in yourmodule.install file to
    manage settings upgrade. You will need to make sure your upgrade will occure
    after the external entity module corresponding one (93010). For example:
    ```
    /**
     * Implements hook_update_dependencies().
     */
    function yourmodule_update_dependencies() {
      $dependencies['yourmodule'][9101] = [
        'external_entities' => 93010,
      ];
      return $dependencies;
    }

    /**
     * Update field mapper configs to new structure.
     */
    function yourmodule_update_9101(&$sandbox) {
      $xntt_storage = \Drupal::entityTypeManager()
        ->getStorage('external_entity_type');
      $xntt_types = $xntt_storage
        ->getQuery()
        ->accessCheck(FALSE)
        ->execute();

      // Loop on external entity types.
      foreach ($xntt_types as $entity_type_id) {
        // Upgrade file storage client ids.
        /** @var \Drupal\Core\Config\Config $config */
        $config = \Drupal::configFactory()->getEditable(
          "external_entities.external_entity_type.{$entity_type_id}"
        );

        $scs = $config->get('storage_clients');
        $need_update = FALSE;
        foreach ($scs as &$sc) {
          if ($sc['id'] == 'yourmodule_pluginid') {
            $need_update = TRUE;
            $sc['config']['performances']['use_index'] =
              $sc['config']['use_index'];
            unset($sc['config']['use_index']);
            $sc['config']['performances']['index_file'] =
              $sc['config']['index_file'];
            unset($sc['config']['index_file']);
            $sc['config']['data_fields']['index_file'] =
              $sc['config']['field_list'];
            unset($sc['config']['field_list']);
            //... do some more stuff if needed.
          }
        }
        if ($need_update) {
          $config->set('storage_clients', $scs);
          $config->save();
        }
      }
    }
    ```

### Upgrading v2 field mappers
@todo Update documentation. v2 field mappers are now v3 property mappers.
!!! NOTE: THE FOLLOWING DOC NEEDS TO BE REVIEWED AND UPDATED !!!
The field mapper approach changed a little: before, each field had a mapping
that could be either an expression or a constant (static) string value. To tell
which type of mapping was used, the field mapper system was using a prefix for
constants. It was not very convenient and was limiting the behavior to 2 cases:
constant and expression.

The new approach gets ride of the prefix and offers instead a drop down for each
field letting the user set if the field is mapped to a constant value or
something else (either an expression or a field name), which brings more
flexibility. Therefore, the previous field mapping value is not a mapping
string anymore but rather an array with 2 keys:
- 'mapping' containing the mapping string and
- 'type' containing the mapping type value as an integer corresponding to
  constants defined in FieldMapperInterface interface
  (src/FieldMapper/FieldMapperInterface.php) as MAPPING_* constants.
Therefore, the field mapper pluging interface must reflect that change and
either provide a way to select the mapping type (dropdown, radios), or provide
a pre-selected mapping type as a hidden value.

For more practical changes:
- In the module's composer.json (if one) update requirements ("require") from
  `"drupal/external_entities": "^2.0@alpha"`
  (or similar) to
  `"drupal/external_entities": "^3.0@alpha"`
- The base class of field mapper plugins is not
  `ConfigurableExpressionFieldMapperBase` anymore but `FieldMapperBase`.
  Replace:
  ```
  use Drupal\external_entities\FieldMapper\ConfigurableExpressionFieldMapperBase;
  ...
  ... extends ConfigurableExpressionFieldMapperBase ...
  ```
  By
  ```
  use Drupal\external_entities\FieldMapper\FieldMapperBase;
  use Drupal\external_entities\FieldMapper\FieldMapperInterface;
  ...
  ... extends FieldMapperBase ...
  ```
  Adding FieldMapperInterface is required to use field mapping type constants.
- Adjust ExternalEntityInterface and ExternalEntityTypeInterface namespaces.
- The default constructor has an addition parameter:
  LoggerChannelFactoryInterface $logger_factory
- Remove the implementation of:
  - ::getConstantMappingPrefix() (not used anymore)
  - ::extractIdFromRawData() (implemented in base class)
  - ::getFieldMappings() (if there are no special needs)
- Add type hinting to methods according to the FieldMapperInterface interface.
- Update ::buildConfigurationForm() according to the changes explained above
  about mapping type selection unless you rely on base implementation.
- ::extractEntityValuesFromRawData(), ::extractFieldValuesFromRawData(), and
  ::extractFieldPropertyValuesFromRawData() method signatures changed, are all
  public and are not using a FieldDefinitionInterface interface anymore.
  You will have to update the logic accordingly and take into account the
  mapping type now. Note: mapped constant values are handled elsewhere so you
  do not have to worry about them in ::extractFieldPropertyValuesFromRawData()
  method.
- Implement ::getPossibleFieldPropertyMappingTypes() that returns the mapping
  types allowed for the given field.
- You may implement ::getMappedFieldName() if some expressions of your field
  mapper can be used to directly map a raw field. For instance, a JSONPath
  expression like '$.somefield' is a direct mapping to the field 'somefield'.
- To log messages, you can now use `$this->logger->notice(...)` or
  `$this->logger->warning(...)` or whatever log method you need.
- Implement a hook_update_N() in yourmodule.install file to manage settings
  upgrade. You will need to make sure your upgrade will occure after the
  external entity module corresponding one (9306). For example:
  ```
  /**
   * Implements hook_update_dependencies().
   */
  function yourmodule_update_dependencies() {
    $dependencies['yourmodule'][9101] = [
      'external_entities' => 9306,
    ];
    return $dependencies;
  }

  /**
   * Update field mapper configs to new structure.
   */
  function yourmodule_update_9101(&$sandbox) {
    $xntt_storage = \Drupal::entityTypeManager()->getStorage('external_entity_type');
    $xntt_types = $xntt_storage
      ->getQuery()
      ->accessCheck(FALSE)
      ->execute();

    // Loop on external entity types.
    foreach ($xntt_types as $entity_type_id) {
      // Upgrade field mapper configs.
      /** @var \Drupal\Core\Config\Config $config */
      $config = \Drupal::configFactory()->getEditable(
        "external_entities.external_entity_type.{$entity_type_id}"
      );

      $fms = $config->get('field_mappers') ?? [];
      $need_upgrade = FALSE;
      foreach ($fms as &$fm) {
        if ('yourmodule_pluginid' == $fm['id']) {
          $need_upgrade = TRUE;
          foreach (
            $fm['config']['field_mappings']
            as $field_name => &$properties
          ) {
            foreach ($properties as $property => &$mapping) {
              // Check for old configs that need upgrade.
              if (is_string($mapping)) {
                $mapping_type = FieldMapperInterface::MAPPING_EXPRESSION;
                if ('+' == $mapping[0]) {
                  $mapping = substr($mapping, 1);
                  $mapping_type = FieldMapperInterface::MAPPING_CONSTANT;
                }
                elseif (<some checks on the $mapping content>) {
                  // You may rewrite $mapping.
                  // And set the mapping type.
                  $mapping_type = FieldMapperInterface::MAPPING_DIRECT;
                }
                //... Add other needed tests.
                $mapping = [
                  'type' => $mapping_type,
                  'mapping' => $mapping,
                ];
              }
            }
          }
        }
      }
      if ($need_upgrade) {
        $config->set('field_mappers', $fms);
        $config->save();
      }
    }
  }
  ```


## Creating a storage client plugin

@todo

## Creating a field mapper plugin

@todo

## Creating a property mapper plugin

@todo

## Creating a data processor plugin

@todo

## Creating a data aggregator plugin

@todo

## Creating external entity types programmatically

@todo To complete.

The constraints should be defined your module install file "hook_install()"
implementation as follow. Don't forget to also clear those changes in your
"hook_uninstall()" implementation.
```
  $locked = \Drupal::state()->get('external_entities.type.locked');
  // Setup constraints on the external entity type 'xnttex1':
  $locked['xnttex1'] = [
    // To disable completly type editing.
    'lock_edit' => FALSE,
    // To disable type removal.
    'lock_delete' => TRUE,
    // To only disable type field editing (add/remove but not display).
    'lock_fields' => TRUE,
    // To disable type field editing and display changes.
    'lock_display' => FALSE,
    // To disable type annotation changes.
    'lock_annotations' => TRUE,
    // To restrict field mapping options.
    'lock_field_mappers' => [
      [
        '<field_name1>' => [
          // To restrict field mapper selection.
          'allow_plugins' => [
            'generic' => 'generic',
            'text' => 'text',
          ],
          // To disable field mapping config changes.
          'lock_config' => TRUE,
          // To completely hide field mapping config.
          'hide_config' => TRUE,
        ],
        '<field_name2>' => [
          'allow_plugins' => [
            // Restrict the use of the 'text 'field mapper to field_name2.
            'text' => 'text',
          ],
        ],
        // For all fields (overrided by field-specific settings).
        '*' => [
          'allow_plugins' => [
            'generic' => 'generic',
          ],
          // To manage default field mapping config changes.
          'lock_config' => FALSE,
          // To manage default visibility of field mapping configs.
          'hide_config' => FALSE,
        ],
      ],
    ],
    'lock_data_aggregator' => [
    // To restrict type data aggregator selection.
      'allow_plugins' => [
        'group' => 'group',
      ],
      // To disable type data aggregator config changes.
      'lock_config' => FALSE,
      // To completely hide type data aggregator config.
      'hide_config' => FALSE,
      // To restrict storage clients.
      'lock_storage_clients' => [
        // Restriction for first storage client (string keys allowed).
        0 => [
          // To restrict type storage client selection.
          'allow_plugins' => [
            'rest' => 'rest',
          ],
          // To disable type storage client config changes.
          'lock_config' => TRUE,
          // To completely hide type storage client config.
          'hide_config' => FALSE,
        ],
      ],
    ],
  ];
  Drupal::state()->set('external_entities.type.locked', $locked);
```
