<?php

namespace Drupal\external_entities\Entity;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\external_entities\DataAggregator\DataAggregatorInterface;
use Drupal\external_entities\DataAggregator\DataAggregatorManager;
use Drupal\external_entities\Event\ExternalEntitiesEvents;
use Drupal\external_entities\Event\ExternalEntityGetMappableFieldsEvent;
use Drupal\external_entities\FieldMapper\FieldMapperInterface;
use Drupal\external_entities\FieldMapper\FieldMapperManager;
use Psr\Log\LoggerInterface;

/**
 * Defines the external_entity_type entity.
 *
 * @ConfigEntityType(
 *   id = "external_entity_type",
 *   label = @Translation("External entity type"),
 *   handlers = {
 *     "list_builder" = "Drupal\external_entities\ExternalEntityTypeListBuilder",
 *     "form" = {
 *       "add" = "Drupal\external_entities\Form\ExternalEntityTypeForm",
 *       "edit" = "Drupal\external_entities\Form\ExternalEntityTypeForm",
 *       "delete" = "Drupal\external_entities\Form\ExternalEntityTypeDeleteForm",
 *     }
 *   },
 *   config_prefix = "external_entity_type",
 *   admin_permission = "administer external entity types",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label",
 *   },
 *   links = {
 *     "edit-form" = "/admin/structure/external-entity-types/{external_entity_type}",
 *     "delete-form" = "/admin/structure/external-entity-types/{external_entity_type}/delete",
 *   },
 *   config_export = {
 *     "id",
 *     "label",
 *     "label_plural",
 *     "description",
 *     "content_class",
 *     "generate_aliases",
 *     "read_only",
 *     "debug_level",
 *     "field_mappers",
 *     "field_mapping_notes",
 *     "data_aggregator",
 *     "data_aggregator_notes",
 *     "persistent_cache_max_age",
 *     "annotation_entity_type_id",
 *     "annotation_bundle_id",
 *     "annotation_field_name",
 *     "inherits_annotation_fields"
 *  }
 * )
 */
class ExternalEntityType extends ConfigEntityBase implements ConfigurableExternalEntityTypeInterface {

  /**
   * Indicates that entities of this external entity type should not be cached.
   */
  const CACHE_DISABLED = 0;

  /**
   * Default field mapper plugin id.
   */
  const DEFAULT_FIELD_MAPPER = 'generic';

  /**
   * Default data aggregator plugin id.
   */
  const DEFAULT_DATA_AGGREGATOR = 'single';

  /**
   * The external entity type ID.
   *
   * @var string
   */
  protected $id;

  /**
   * The human-readable name of the external entity type.
   *
   * @var string
   */
  protected $label;

  /**
   * The plural human-readable name of the external entity type.
   *
   * @var string
   */
  protected $label_plural;

  /**
   * The external entity type description.
   *
   * @var string
   */
  protected $description;

  /**
   * The external entity class used for content instances.
   *
   * @var string
   */
  protected $content_class = ExternalEntity::class;

  /**
   * Whether or not entity types of this external entity type are read only.
   *
   * @var bool
   */
  protected $read_only;

  /**
   * Debug level of current instance.
   *
   * @var int
   */
  protected $debug_level;

  /**
   * Whether or not to auto-generate aliases for this external entity type.
   *
   * @var bool
   */
  protected $generate_aliases;

  /**
   * Array of field mapper plugin settings.
   *
   * @var array
   */
  protected $field_mappers = [];

  /**
   * Field mapping administrative notes.
   *
   * @var string
   */
  protected $field_mapping_notes = '';

  /**
   * Array a mappable fields of this instance.
   *
   * @var \Drupal\Core\Field\FieldDefinitionInterface[]
   */
  protected $mappableFields = [];

  /**
   * The field mapper plugin instance.
   *
   * @var \Drupal\external_entities\FieldMapper\FieldMapperInterface
   */
  protected $fieldMapperPlugins = [];

  /**
   * Data aggregator plugin settings.
   *
   * @var array
   */
  protected $data_aggregator = [];

  /**
   * Data aggregator plugin instance.
   *
   * @var \Drupal\external_entities\DataAggregator\DataAggregatorInterface
   */
  protected $dataAggregator;

  /**
   * Data aggregator administrative notes.
   *
   * @var string
   */
  protected $data_aggregator_notes = '';

  /**
   * Max age entities of this external entity type may be persistently cached.
   *
   * @var int
   */
  protected $persistent_cache_max_age = self::CACHE_DISABLED;

  /**
   * The annotations entity type id.
   *
   * @var string
   */
  protected $annotation_entity_type_id;

  /**
   * The annotations bundle id.
   *
   * @var string
   */
  protected $annotation_bundle_id;

  /**
   * The field this external entity is referenced from by the annotation entity.
   *
   * @var string
   */
  protected $annotation_field_name;

  /**
   * Local cache for the annotation field.
   *
   * @var array
   *
   * @see ExternalEntityType::getAnnotationField()
   */
  protected $annotationField;

  /**
   * Indicates if the external entity inherits the annotation entity fields.
   *
   * @var bool
   */
  protected $inherits_annotation_fields = FALSE;

  /**
   * {@inheritdoc}
   */
  public function createDuplicate() {
    $clone = parent::createDuplicate();
    // Clear local plugin instances to make sure they won't be shared.
    $clone->fieldMapperPlugins = [];
    $clone->dataAggregator = NULL;
    return $clone;
  }

  /**
   * {@inheritdoc}
   */
  public function getLogger() :LoggerChannel {
    return $this->logger($this->id() ?: 'external_entities');
  }

  /**
   * {@inheritdoc}
   */
  public function getLabel() :string {
    return $this->label ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function getPluralLabel() :string {
    return $this->label_plural ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() :string {
    return $this->description ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function isReadOnly() :bool {
    return (bool) $this->read_only;
  }

  /**
   * {@inheritdoc}
   */
  public function getDebugLevel() :int {
    return $this->debug_level ?? 0;
  }

  /**
   * {@inheritdoc}
   */
  public function setDebugLevel(int $debug_level = 1) {
    $this->debug_level = $debug_level;
  }

  /**
   * {@inheritdoc}
   */
  public function automaticallyGenerateAliases() :bool {
    return (bool) $this->generate_aliases;
  }

  /**
   * {@inheritdoc}
   */
  public function getMappableFields(bool $reload = FALSE) :array {
    if (empty($this->mappableFields) || $reload) {
      $this->mappableFields = [];

      $derived_entity_type = $this->getDerivedEntityType();

      $fields = [];
      if (!empty($derived_entity_type)) {
        $fields = $this
          ->entityFieldManager()
          ->getFieldDefinitions(
            $derived_entity_type->id(),
            $derived_entity_type->id()
          );
      }
      $this->mappableFields = array_filter(
        $fields, function (FieldDefinitionInterface $field) {
          return $field->getName() !== ExternalEntityInterface::ANNOTATION_FIELD
            && !$field->isComputed();
        }
      );

      // Allow other modules to alter mappable fields.
      $event = new ExternalEntityGetMappableFieldsEvent(
        $this,
        $this->mappableFields
      );
      \Drupal::service('event_dispatcher')->dispatch(
        $event,
        ExternalEntitiesEvents::GET_MAPPABLE_FIELDS
      );

      $this->mappableFields = $event->getMappableFields();
    }

    return $this->mappableFields;
  }

  /**
   * {@inheritdoc}
   */
  public function getRequiredFields() :array {
    $fields = [
      'id',
      'title',
    ];

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMapperDefaultConfiguration(
    string $field_name
  ) :array {
    return [
      // Allow the mapper to call back into the entity type (e.g., to fetch
      // additional data like field lists from the remote service).
      ExternalEntityTypeInterface::XNTT_TYPE_PROP => $this,
      'field_name' => $field_name,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMapperId(string $field_name) :string {
    if (!empty($this->fieldMapperPlugins[$field_name])) {
      $this->field_mappers[$field_name]['id'] =
        $this->fieldMapperPlugins[$field_name]->getPluginId();
    }
    return $this->field_mappers[$field_name]['id'] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function setFieldMapperId(
    string $field_name,
    string $field_mapper_id
  ) :self {
    // Make sure field is mappable.
    $mappable_fields = $this->getMappableFields();
    if (!empty($field_name) && !empty($mappable_fields[$field_name])) {
      // Only update if changed.
      if ($field_mapper_id != ($this->field_mappers[$field_name]['id'] ?? '')) {
        // Initialize the field mapper setting array if not already.
        $this->field_mappers[$field_name] ??= ['config' => []];
        // Set corresponding  the field mapper plugin.
        $this->field_mappers[$field_name]['id'] = $field_mapper_id;
        // Remove corresponding current field mapper instance.
        $this->fieldMapperPlugins[$field_name] = NULL;
      }
    }
    else {
      $this->getLogger()->warning(
        'Tried to set a field mapper on a non-mappable field ('
        . $field_name
        . ').'
      );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMapper(
    string $field_name
  ) :FieldMapperInterface|null {
    // Make sure field is mappable.
    $mappable_fields = $this->getMappableFields();
    if (empty($field_name)) {
      $this->getLogger()->warning(
        'No field machine name provided!'
      );
      return NULL;
    }
    elseif (empty($mappable_fields[$field_name])) {
      $this->getLogger()->warning(
        "Given field ($field_name) is not mappable!"
      );
      return NULL;
    }
    elseif (empty($this->field_mappers[$field_name]['id'])) {
      // Not mapped.
      return NULL;
    }

    if (empty($this->fieldMapperPlugins[$field_name])) {
      $field_mapper_plugin_manager =
        \Drupal::service('plugin.manager.external_entities.field_mapper');
      assert($field_mapper_plugin_manager instanceof FieldMapperManager);

      $config = NestedArray::mergeDeep(
        $this->getFieldMapperDefaultConfiguration($field_name),
        $this->getFieldMapperConfig($field_name)
      );

      try {
        $this->fieldMapperPlugins[$field_name] =
          $field_mapper_plugin_manager->createInstance(
            $this->getFieldMapperId($field_name),
            $config
          );
        // Adjust field mapper debug level to current one only if set.
        if ($debug_level = $this->getDebugLevel()) {
          $this->fieldMapperPlugins[$field_name]->setDebugLevel($debug_level);
        }
      }
      catch (PluginException $e) {
        $this->getLogger()->warning(
          "Failed to load field mapper pluggin!\n"
          . $e
        );
        return NULL;
      }

      // Adjust field mapper debug level to current one only if set.
      if ($debug_level = $this->getDebugLevel()) {
        $this->fieldMapperPlugins[$field_name]->setDebugLevel($debug_level);
      }
    }

    return $this->fieldMapperPlugins[$field_name];
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMapperConfig(string $field_name) :array {
    if (!empty($this->fieldMapperPlugins[$field_name])) {
      $this->field_mappers[$field_name]['config'] =
        $this->fieldMapperPlugins[$field_name]->getConfiguration();
    }
    return $this->field_mappers[$field_name]['config'] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function setFieldMapperConfig(
    string $field_name,
    array $field_mapper_config
  ) :self {
    // Make sure field is mappable.
    $mappable_fields = $this->getMappableFields();
    if (!empty($field_name)
      && !empty($mappable_fields[$field_name])
      && !empty($this->field_mappers[$field_name]['id'])
    ) {
      // Update plugin and local config.
      if (!empty($this->fieldMapperPlugins[$field_name])) {
        $this->fieldMapperPlugins[$field_name]
          ->setConfiguration(
            NestedArray::mergeDeep(
              $this->getFieldMapperDefaultConfiguration($field_name),
              $field_mapper_config
            )
          );
      }
      else {
        $this->field_mappers[$field_name]['config'] = $field_mapper_config;
      }
    }
    elseif (empty($this->field_mappers[$field_name]['id'])) {
      $this->getLogger()->warning(
        'Tried to configure mapping on a field ('
        . $field_name
        . ') that has no field mapper set.'
      );
    }
    else {
      $this->getLogger()->warning(
        'Tried to configure mapping on a non-mappable field ('
        . $field_name
        . ').'
      );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldMappingNotes() :string {
    return $this->field_mapping_notes ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function setFieldMappingNotes(
    string $field_mapping_notes
  ) :self {
    $this->field_mapping_notes = $field_mapping_notes;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDataAggregatorDefaultConfiguration() :array {
    return [
      // Allow the aggregator to call back into the entity type.
      ExternalEntityTypeInterface::XNTT_TYPE_PROP => $this,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getDataAggregatorId() :string {
    if (!empty($this->dataAggregator)) {
      $this->data_aggregator['id'] =
        $this->dataAggregator->getPluginId();
    }
    return $this->data_aggregator['id'] ?? static::DEFAULT_DATA_AGGREGATOR;
  }

  /**
   * {@inheritdoc}
   */
  public function setDataAggregatorId(
    string $data_aggregator_id,
  ) :self {
    // Only update if changed.
    if ($data_aggregator_id != ($this->data_aggregator['id'] ?? '')) {
      $this->data_aggregator['id'] = $data_aggregator_id;
      $this->data_aggregator['config'] = [];
      $this->dataAggregator = NULL;
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDataAggregator() :DataAggregatorInterface {

    if (empty($this->dataAggregator)) {
      $data_aggregator_plugin_manager =
        \Drupal::service('plugin.manager.external_entities.data_aggregator');
      assert($data_aggregator_plugin_manager instanceof DataAggregatorManager);

      $aggregator_id = $this->getDataAggregatorId();
      $config = NestedArray::mergeDeep(
        $this->getDataAggregatorDefaultConfiguration(),
        $this->getDataAggregatorConfig()
      );

      $this->dataAggregator = $data_aggregator_plugin_manager->createInstance(
        $aggregator_id,
        $config
      );

      // Adjust data aggregator debug level to current one only if set.
      if ($debug_level = $this->getDebugLevel()) {
        $this->dataAggregator->setDebugLevel($debug_level);
      }
    }
    return $this->dataAggregator;
  }

  /**
   * {@inheritdoc}
   */
  public function getDataAggregatorConfig() :array {
    if (!empty($this->dataAggregator)) {
      $this->data_aggregator['config'] =
        $this->dataAggregator->getConfiguration();
    }
    return $this->data_aggregator['config'] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function setDataAggregatorConfig(
    array $data_aggregator_config
  ) :self {
    if (!empty($this->data_aggregator['id'])) {
      if (!empty($this->dataAggregator)) {
        // Update plugin and local config.
        $this->dataAggregator->setConfiguration(
            NestedArray::mergeDeep(
              $this->getDataAggregatorDefaultConfiguration(),
              $data_aggregator_config
            )
          );
        $this->data_aggregator['config'] =
          $this->dataAggregator->getConfiguration();
      }
      else {
        $this->data_aggregator['config'] = $data_aggregator_config;
      }
    }
    else {
      $this->getLogger()->warning(
        'Tried to set data aggregator config with no data aggregator set.'
      );
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDataAggregatorNotes() :string {
    return $this->data_aggregator_notes ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function setDataAggregatorNotes(
    string $data_aggregator_notes
  ) :self {
    $this->data_aggregator_notes = $data_aggregator_notes;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getPersistentCacheMaxAge() :int {
    return $this->persistent_cache_max_age;
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(EntityStorageInterface $storage, $update = TRUE) {
    parent::postSave($storage, $update);

    // Clear the entity type definitions cache so changes flow through to the
    // related entity types.
    $this->entityTypeManager()->clearCachedDefinitions();

    // Clear the router cache to prevent RouteNotFoundException errors caused
    // by the Field UI module.
    \Drupal::service('router.builder')->rebuild();

    // Rebuild local actions so that the 'Add field' action on the 'Manage
    // fields' tab appears.
    \Drupal::service('plugin.manager.menu.local_action')->clearCachedDefinitions();

    // Clear the static and persistent cache.
    $storage->resetCache();
    if ($this->entityTypeManager()->hasDefinition($this->id())) {
      $this
        ->entityTypeManager()
        ->getStorage($this->id())
        ->resetCache();
    }

    $edit_link = $this->toLink(t('Edit entity type'), 'edit-form')->toString();
    if ($update) {
      $this->getLogger()->notice(
        'Entity type %label has been updated.',
        ['%label' => $this->label(), 'link' => $edit_link]
      );
    }
    else {
      // Notify storage to create the database schema.
      $entity_type = $this->entityTypeManager()->getDefinition($this->id());
      \Drupal::service('entity_type.listener')
        ->onEntityTypeCreate($entity_type);

      $this->getLogger()->notice(
        'Entity type %label has been added.',
        ['%label' => $this->label(), 'link' => $edit_link]
      );
    }

  }

  /**
   * {@inheritdoc}
   */
  public static function postDelete(EntityStorageInterface $storage, array $entities) {
    parent::postDelete($storage, $entities);
    \Drupal::service('entity_type.manager')->clearCachedDefinitions();
  }

  /**
   * {@inheritdoc}
   */
  public function getDerivedEntityTypeId() :string {
    return $this->id() ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function getDerivedEntityType() :ContentEntityTypeInterface|null {
    return $this->entityTypeManager()->getDefinition($this->getDerivedEntityTypeId(), FALSE);
  }

  /**
   * {@inheritdoc}
   */
  public function isAnnotatable() :bool {
    return $this->getAnnotationEntityTypeId()
      && $this->getAnnotationBundleId()
      && $this->getAnnotationFieldName();
  }

  /**
   * {@inheritdoc}
   */
  public function getAnnotationEntityTypeId() :string|null {
    return $this->annotation_entity_type_id;
  }

  /**
   * {@inheritdoc}
   */
  public function getAnnotationBundleId() :string|null {
    return $this->annotation_bundle_id ?: $this->getAnnotationEntityTypeId();
  }

  /**
   * {@inheritdoc}
   */
  public function getAnnotationFieldName() :string|null {
    return $this->annotation_field_name;
  }

  /**
   * Returns the entity field manager.
   *
   * @return \Drupal\Core\Entity\EntityFieldManagerInterface
   *   The entity field manager.
   */
  protected function entityFieldManager() :EntityFieldManagerInterface {
    return \Drupal::service('entity_field.manager');
  }

  /**
   * {@inheritdoc}
   */
  public function getAnnotationField() :FieldDefinitionInterface|null {
    if (!isset($this->annotationField) && $this->isAnnotatable()) {
      $field_definitions = $this->entityFieldManager()->getFieldDefinitions($this->getAnnotationEntityTypeId(), $this->getAnnotationBundleId());
      $annotation_field_name = $this->getAnnotationFieldName();
      if (!empty($field_definitions[$annotation_field_name])) {
        $this->annotationField = $field_definitions[$annotation_field_name];
      }
    }

    return $this->annotationField;
  }

  /**
   * {@inheritdoc}
   */
  public function inheritsAnnotationFields() :bool {
    return (bool) $this->inherits_annotation_fields;
  }

  /**
   * {@inheritdoc}
   */
  public function getBasePath() :string {
    return str_replace('_', '-', strtolower($this->id));
  }

  /**
   * Gets the logger for a specific channel.
   *
   * @param string $channel
   *   The name of the channel.
   *
   * @return \Psr\Log\LoggerInterface
   *   The logger for this channel.
   */
  protected function logger($channel) :LoggerInterface {
    return \Drupal::getContainer()->get('logger.factory')->get($channel);
  }

  /**
   * {@inheritdoc}
   */
  public function get($property_name) {
    // Get "local" config data from plugins themselves as it may have changed.
    if ('field_mappers' == $property_name) {
      // Only update where we got plugins that may have altered config.
      foreach ($this->fieldMapperPlugins as $field_name => $field_mapper) {
        if (!empty($field_mapper)) {
          $this->field_mappers[$field_name] = [
            'id' => $field_mapper->getPluginId(),
            'config' => $field_mapper->getConfiguration(),
          ];
        }
      }
    }
    elseif ('data_aggregator' == $property_name) {
      // Get current data aggregator config.
      if (!empty($this->dataAggregator)) {
        $this->data_aggregator = [
          'id' => $this->getDataAggregatorId(),
          'config' => $this->getDataAggregatorConfig(),
        ];
      }
    }
    return parent::get($property_name);
  }

  /**
   * {@inheritdoc}
   */
  public function __sleep() {
    // Prevent some properties from being serialized.
    return array_diff(parent::__sleep(), [
      'fieldMapperPlugins',
      'dataAggregator',
      'annotationField',
    ]);
  }

}
